# -*- coding: utf-8 -*-
"""
Created on Mon Nov  9 15:43:25 2020

@author: Jean

@inproceedings{he-choi-2021-stem,
    title = "The Stem Cell Hypothesis: Dilemma behind Multi-Task Learning with Transformer Encoders",
    author = "He, Han and Choi, Jinho D.",
    booktitle = "Proceedings of the 2021 Conference on Empirical Methods in Natural Language Processing",
    month = nov,
    year = "2021",
    address = "Online and Punta Cana, Dominican Republic",
    publisher = "Association for Computational Linguistics",
    url = "https://aclanthology.org/2021.emnlp-main.451",
    pages = "5555--5577",
    abstract = "Multi-task learning with transformer encoders (MTL) has emerged as a powerful technique to improve performance on closely-related tasks for both accuracy and efficiency while a question still remains whether or not it would perform as well on tasks that are distinct in nature. We first present MTL results on five NLP tasks, POS, NER, DEP, CON, and SRL, and depict its deficiency over single-task learning. We then conduct an extensive pruning analysis to show that a certain set of attention heads get claimed by most tasks during MTL, who interfere with one another to fine-tune those heads for their own objectives. Based on this finding, we propose the Stem Cell Hypothesis to reveal the existence of attention heads naturally talented for many tasks that cannot be jointly trained to create adequate embeddings for all of those tasks. Finally, we design novel parameter-free probes to justify our hypothesis and demonstrate how attention heads are transformed across the five tasks during MTL through label analysis.",
}

"""
# Upgraded to HanLP2.1  2022/05/13
# 记得为hanlp_trie.trie.py增加 parse_longest2()函数。

import time, math, random
import pandas as pd
import collections
import json,re
from igraph import *
import hanlp
# from hanlp.common.trie import Trie  # HanLP2.0 -->2.1
from hanlp_trie.trie import Trie


# 句法依存分析前对汇总后的货物劳务名称做预处理
def preprocess(row):     
    text = row["hwmc"]
    # 去掉前后空格等特殊处理
    text = text.strip()
    if text == "." or text == "/"  or text == "-":
        text = ""    
    # 去掉几种括号及其中内容，如果去掉后没有内容了，则予以保留
    text1 = text
    text = re.sub("\\(.*?\\)|（.*?）", "", text)
    if len(text) == 0:
        text = text1
    text1 = text
    text = re.sub("\\[.*?\\]|【.*?】", "", text)
    if len(text) == 0:
        text = text1
    text1 = text
    text = re.sub("\\{.*?\\}|｛.*?｝", "", text)
    if len(text) == 0:
        text = text1
    # 去掉商品及货物分类简称
    text1 = text        
    text = re.sub("\\*.*?\\*", "", text)
    if len(text) == 0:
        text = text1
    # 去掉单边括号后的内容，这种情况不少
    text1 = text        
    text = text.split("（")[0]  #	722 维达（Vinda 	822	荣耀8全网通（FRD 	825	彩电挂式安装调试卡（
    if len(text) == 0:
        text = text1
    text1 = text                
    text = text.split("）")[0] 
    if len(text) == 0:
        text = text1  
    text1 = text                
    text = text.split("(")[0]  
    if len(text) == 0:
        text = text1  
    text1 = text                
    text = text.split(")")[0]  
    if len(text) == 0:
        text = text1  
    text1 = text
    text = re.sub("及", " ", text)  # 国际快件费及附加 -> 国际快件费 附加
    if len(text) == 0:
        text = text1        
    text1 = text                
    text = text.split(" ")[0]  # 空格分隔的多个并列，只留第一个，如 "小米5C 移动版 3"
    if len(text) == 0:
        text = text1        
    text = re.sub("[0-9]L", "", text)  # 胶瓶美汁源果粒橙12*1.25L，去掉L，引起分词错误 
    
    return text


# 更改分割函数，检查一下取词是否输出空串, 抛弃trie树返回的结果中与前一个词有重叠的词
def split_sents(text: str, trie: Trie):
    #words = trie.parse_longest(text)     # HandLP2.0 -->2.1
    words = trie.parse_longest2(text)    
    sents = []
    words2 = []
    pre_start = 0
    offsets = []
    for word, value, start, end in words:
        # print(word, value, start, end, pre_start, start)
        if pre_start != start:
            word2 = text[pre_start: start]
            if len(word2)>0:
                # print(word2)
                sents.append(word2)
                words2.append(tuple([word,value,start,end]))
                offsets.append(pre_start)
                pre_start = end
            else:
                # print("自定义词有重叠，跳过:",word,value,start,end)
                pass
        else:
            # print("自定义词首尾相接：",pre_start, start)
            words2.append(tuple([word,value,start,end]))
            pre_start = end

    if pre_start != len(text):
        sents.append(text[pre_start:])
        offsets.append(pre_start)
    return sents, offsets, words2


# 合并输入为单一句子，以便用自定义词典分割后用GPU并行处理， “~~”是句子分隔符，用于后面重新分割句子。
def merge_sents(hlhzs):
    return "~~".join(hlhzs)


# 自定义函数，再合并自定义分词及词性标注，剩余部分的词性分词、词性标注
def merge_parts(tokens, postags, offsets, words):
    tps = []
    for i in range(len(tokens)):
        tps2 = [(token, postag) for (token, postag) in zip(tokens[i], postags[i])]
        tps.append(tps2)
    items = [(i, ts) for (i, ts) in zip(offsets, tps)]                
    items += [(start, [(word, value)]) for (word, value, start, end) in words]
    return [each for x in sorted(items) for each in x[1]]


# 根据句子分隔符重新按句子分割分词与词性标注的结果，以便用GPU并行作语法和语义分析。
def split_pos(postags):
    tpsents = []; tpsent = []
    for postag in postags:
        if postag[0] == "~~":
            tpsents.append(tpsent)
            tpsent = []
        else:
            tpsent.append(postag)        
    tpsents.append(tpsent)
    return tpsents


# 建立语法树（图）/语义网络（图）函数， 语义分析结果为一个有向无环图，称为语义依存图（Semantic Dependency Graph）。
def getTree(i,text):
    res = json.loads(text)
    try:
        items = res["items"]
    except KeyError:
        print(i,text)
        return None
    ids = []; pos = []; words = []; head = []; deprel = []
    for item in items:
        ids.append(item["id"]); words.append(item["form"]); 
        # pos.append(item["cpos"])     # HanLP2.0 -->2.1
        try:
            pos.append(item["upos"])   # 语法树（图）parser1,3,4,多数
        except KeyError:
            pos.append(item["cpos"])   # 语义树（图）parser2,少数
        head.append(item["head"]); deprel.append(item["deprel"])
    vs = pd.DataFrame({"id":ids,"word":words,"postag":pos})
    # 存在一对多的情况，每个对应关系展开成一条边。 head与deprel为list。
    try: 
        ids2 = []; head2 = []; deprel2 = []
        for i in range(len(ids)):
            for j in range(len(head[i])):
                ids2.append(ids[i]); head2.append(head[i][j]); deprel2.append(deprel[i][j])
        es = pd.DataFrame({"id":ids2,"head":head2,"deprel":deprel2})
    # 没有一对多的情况，不展开
    except Exception:
        es = pd.DataFrame({"id":ids,"head":head,"deprel":deprel})
    vd = vs.to_dict(orient='records')
    ed = es.to_dict(orient='records')
    g = Graph.DictList(vd,ed, vertex_name_attr="id",edge_foreign_keys=('id', 'head'),directed=True)
    # 给根节点赋非空值
    vs = g.vs.select(id=0)   
    vs["word"] = ""; vs["postag"] = ""
    g.vs["label"]=[str(v.index)+":"+v["word"]+"\n"+v["postag"] for v in g.vs]
    g.es["label"]=g.es["deprel"]
    return g


# 从根结点开始遍历语法树，取最先出现的名词作货劳名称，所以层级高的优先，同层级的在后面的优先
def get_hlmc(g, verbose = False):
    # 建立邻接表
    adj = [[n.index for n in v.neighbors("in")] for v in g.vs]
    spcialWords = re.compile("[\.|\-|%]") #特殊名称， "."软件版本号， "-"具体型号，跳过
    spcialPuncs = ["-", "/", "*", "+"]       #特殊标点符号，标示里面包含并列关系,
    skipWords = re.compile("[顶|装]")
    # 广度优先搜索遍历语法树
    def traverse(adj, paths, depth):
        # 记录遍历的深度
        depth += 1;  nxt_paths = []
        # 遍历所有的路径
        for path in paths['paths']:
            v = g.vs[path[-1]]
            # 如果不是并列词，则反转邻接表，改成同层级的后面的优先，排在后面的名词优先，
            # 发票语法结构的特点，前面的名词多数是修饰语         
            if v["word"] not in spcialPuncs and depth>1:
                adj[path[-1]].reverse()
            for idx in adj[path[-1]]:
                v1 = g.vs[idx]
                # 如 汤臣倍健多种维生素咀嚼片-迪士尼漫威装, 根结点是 漫威装
                if v1["word"] in spcialPuncs:
                    adj[path[-1]].reverse()  # 子结点中有并列关系，邻接表恢复为正常顺序。                    
                    break
            # 遍历路径中最后一个结点的所有邻居
            for nxt in adj[path[-1]]:
                v = g.vs[nxt]
                if verbose:
                    print(depth,v["word"],v["postag"])
                if "N"  in v["postag"]:   # 如果是名词，则加入候选货劳清单中
                    flag = True
                    if spcialWords.findall(v["word"]) and v["postag"]!= "NM" :  # 软件版本号，具体型号，跳过，药名除外。
                        flag = False
                    if v["postag"]== "NT":  # 品牌，跳过
                        flag = False
                    matched = skipWords.search(v["word"])   # 检查是否包含要跳过的特殊名词
                    if matched:
                        if matched.span()[0]+1 == len(v["word"]):   # 漫威装 在最后，跳过；安装费 在中间，接受
                            flag = False
                    if flag:
                        paths["hlmc"].append(v["word"])
                        paths["pos"].append(v["postag"])
                nxt_path = path + [nxt]   # 在路径中加入下一个结点，生成新的路径
                nxt_paths.append(nxt_path)
        # 路径的深度增加了一级
        paths['paths'] = nxt_paths                            
        if len(nxt_paths)==0:
            # 没有新路径了
            return paths
        else:
            # 往路径深度递归调用一级
            return traverse(adj, paths, depth)
    # 从根结点开始递归遍历语法树
    v = g.vs.select(id=0)[0]
    # 返回候选货劳名称列表，排在前面的在语法树中的层级较高，同层级中则排在较后的优先级较高
    res = traverse(adj, {'paths': [[v.index]], 'hlmc': [], "pos": []}, 0)
    # 没有找到名词，返回一级结点作货劳名称
    if len(res["hlmc"]) == 0:
        v1 = v.neighbors("in")[0]
        res["hlmc"].append(v1["word"]); res["pos"].append(v1["postag"])
    
    return res
    

# 调用GPU，并行处理，分词、词性标注、语义依存分析，提取货物劳务名称
def ParseGPU(parser, hlhz, flag):
    # 调用GPU，并行处理，分词、词性标注、语义依存分析
    res = parser(list(hlhz["hwmc2"]))

    # 从GPU处理结果中提取货物劳务名称
    restr = []; reshl = []; pos = []; method = []
    # for i in range(len(res.dependencies)):               # HanLP2.0 -->2.1
    for i in range(len(res["dependencies"])):
        # items = res.dependencies[i]        # HanLP2.0 -->2.1
        items = res["dependencies"][i]
        items2 = {}; items2.setdefault("items",[])
        #                  if key in ["cpos","deprel","form","head","id"]}         # HanLP2.0 -->2.1        
        for item in items:
            item2 = {key: value for key, value in item.items() \
                           if key in ["upos","cpos","deprel","form","head","id"]}  # HanLP2.0 -->2.1
            items2.get("items").append(item2)
        seStr = json.dumps(items2,ensure_ascii=False)
        try:
            g = getTree(i,seStr)     # 转换成语义图 
            #random.seed(42)
            #plot(g,bbox=(400,300))
            hlmc = get_hlmc(g, False) # 遍历语义图得到货物劳务名称
            #print(hlmc)
            reshl.append(hlmc["hlmc"][0]); pos.append(hlmc["pos"][0])
            method.append(flag); restr.append(seStr)
        except Exception:   #图上形成了循环，没有 id=0的根结点，引起出错。
            reshl.append(None); pos.append(None); restr.append(None); method.append(None)            
    #合并进汇总货劳名称子集
    #print(len(reshl),len(pos),len(restr),len(method))
    hlhz["hlmc"] = reshl; hlhz["pos"] = pos; hlhz["reStr"] = restr; hlhz["method"] = method
    return hlhz



# 装入用户自定义词典
print("加载用户自定义词典...")
t1 = time.time()
types = {"word":str,"pos":str}
dicZdy = pd.read_csv('D:/temp/data/增值税自开票2017/词汇导入demo-HanLP.csv', sep=',',\
                     header=None, dtype=types, encoding="GBK")
# dicZdy.columns = ["word","pos"]
dicZdy = dict(collections.OrderedDict(zip(dicZdy.iloc[:,0],dicZdy.iloc[:,1])))
# 挂载自定义词典
trie = Trie()
# key是自定义词， value可以是词性等，由业务定义。
trie.update(dicZdy)
t2 = time.time()
print(t2-t1)


# 装入HanLP使用的分词、词性标注、依存句法分析等各种模型, 96秒，有点慢，模型数据太庞大了。
t1 = time.time()
# 装入分词模型
print("装入HanLP分词、词性标注、依存句法分析、语义依存分析模型...")
#tokenizer = hanlp.load('LARGE_ALBERT_BASE')
tokenizer = hanlp.load(hanlp.pretrained.tok.LARGE_ALBERT_BASE)
# 装入词性标注模型
tagger = hanlp.load(hanlp.pretrained.pos.CTB5_POS_RNN_FASTTEXT_ZH)
# 装入依存句法分析模型
syntactic_parser = hanlp.load(hanlp.pretrained.dep.CTB7_BIAFFINE_DEP_ZH)
# 装入语义依存分析模型
semantic_parser = hanlp.load(hanlp.pretrained.sdp.SEMEVAL16_NEWS_BIAFFINE_ZH)
# 定义流水线，不包含自定义词库引用
# 依存句法分析
parser1 = hanlp.pipeline() \
    .append(tokenizer, output_key='tokens') \
    .append(tagger,input_key='tokens', output_key='postags') \
    .append(syntactic_parser, input_key=('tokens','postags'), output_key='dependencies')    
# 语义依存分析
parser2 = hanlp.pipeline() \
    .append(tokenizer, output_key='tokens') \
    .append(tagger,input_key='tokens', output_key='postags') \
    .append(semantic_parser, input_key=('tokens','postags'), output_key='dependencies')
# 定义流水线，引用自定义词库，分词、词性标注、合并、依存句法分析/语义依存分析
parser3 = hanlp.pipeline() \
    .append(merge_sents, output_key="sents" ) \
    .append(split_sents, input_key="sents", output_key=('parts', 'offsets', 'words'), trie=trie) \
    .append(tokenizer, input_key='parts', output_key='tokens') \
    .append(tagger,input_key='tokens', output_key='postags') \
    .append(merge_parts, input_key=('tokens', 'postags', 'offsets', 'words'), output_key='merged') \
    .append(split_pos, input_key="merged", output_key="splits") \
    .append(syntactic_parser, input_key='splits', output_key='dependencies') 
parser4 = hanlp.pipeline() \
    .append(merge_sents, output_key="sents" ) \
    .append(split_sents, input_key="sents", output_key=('parts', 'offsets', 'words'), trie=trie) \
    .append(tokenizer, input_key='parts', output_key='tokens') \
    .append(tagger,input_key='tokens', output_key='postags') \
    .append(merge_parts, input_key=('tokens', 'postags', 'offsets', 'words'), output_key='merged') \
    .append(split_pos, input_key="merged", output_key="splits") \
    .append(syntactic_parser, input_key='splits', output_key='dependencies')     
t2 = time.time()
print("模型已装入...", t2-t1)


# 批量分析处理汇总货物劳务名称
# 重新读入货物劳务汇总
print("装入汇总货劳名称...")
t1 = time.time()
types = {"hwmc":str,"count":int,"je":float}
hlhz = pd.read_csv("D:/temp/data/增值税自开票2017/hlhz-gbk-1k.csv", sep='~',dtype=types, \
                   encoding="GBK",header=None)
hlhz.columns = ["hwmc","count","je"]
# 统计，透视表需要有 index列，增加 flag列以便作全表统计
hlhz["flag"] = 1
hltj = pd.pivot_table(hlhz, index=["flag"], values=['count',"je"],aggfunc=["sum"]).reset_index()
# 每个子集的大小及划分的子集数，先取频数最大的前1万行测试
chunk  = 1000
blocks = math.ceil(len(hlhz["hwmc"])/chunk)
t2 = time.time()
print(t2-t1)

# 第一个子集
print("预处理汇总货劳名称...")
t1 = time.time()
hlhz1 = hlhz[:chunk]
# 预处理货劳名称，以便后面更准确的做依存句法分析，这里先用频数最高的前1万行测试。
hlhz1["hwmc2"] = hlhz1.apply(lambda x:preprocess(x), axis=1)

# 为批处理分开数据成4个子集，是否有自定义词，是否并列式
# "-" 并列式，先用语义分析尝试，效果要好一点
Re1 = re.compile("-")
Re4 = re.compile(r"[1-9a-zA-Z][-][1-9a-zA-Z]")
res1 = []; res2 =[]; res3 = []; res4 =[]
for i in range(hlhz1.shape[0]):
    sents, offsets, words = split_sents(hlhz1["hwmc2"].iloc[i], trie)
    if len(words) == 0: # 没有自定义词
        # 并列式
        if Re1.search(hlhz1["hwmc2"].iloc[i]) and not Re4.search(hlhz1["hwmc2"].iloc[i]):
            hlhz1["hwmc2"].iloc[i] = re.sub("-", "*", hlhz1["hwmc2"].iloc[i])  # 替换"-"为"*"，以帮助正确分词
            res2.append(i)
        else:  # 非并列式
            res1.append(i)            
    else:  # 有自定义词
        # 并列式        
        if Re1.search(hlhz1["hwmc2"].iloc[i]) and not Re4.search(hlhz1["hwmc2"].iloc[i]):
            hlhz1["hwmc2"].iloc[i] = re.sub("-", "*", hlhz1["hwmc2"].iloc[i])  # 替换"-"为"*"，以帮助正确分词            
            res4.append(i)
        else:  # 非并列式
            res3.append(i)            
hlhz2 = hlhz1.iloc[res1]; hlhz2 = hlhz2.reset_index(drop=True)
hlhz3 = hlhz1.iloc[res2]; hlhz3 = hlhz3.reset_index(drop=True)
hlhz4 = hlhz1.iloc[res3]; hlhz4 = hlhz4.reset_index(drop=True)
hlhz5 = hlhz1.iloc[res4]; hlhz5 = hlhz5.reset_index(drop=True)
t2 = time.time()
print(t2-t1)

t1 = time.time()
# 用户自定义词
print("GPU:用户自定义词，非并列...")
# 提取货劳名称
hlhz4 = ParseGPU(parser3, hlhz4, 2)
t2 = time.time()
print(t2-t1)  
# 提取货劳名称
print("GPU:用户自定义词，并列...")
hlhz5 = ParseGPU(parser4, hlhz5, 3)
t2 = time.time()
print(t2-t1)  
# 无用户自定义词
print("GPU:无用户自定义词，非并列...")
hlhz2 = ParseGPU(parser1, hlhz2, 0)
t2 = time.time()
print(t2-t1) 
print("GPU:无用户自定义词，并列...")
hlhz3 = ParseGPU(parser2, hlhz3, 1)
t2 = time.time()
print(t2-t1)  
# 合并识别结果
print("合并处理结果...")
hlhz1 = pd.concat([hlhz2,hlhz3,hlhz4,hlhz5])
hlhz1 = hlhz1.sort_values(["count"], ascending=False)
hlhz1 = hlhz1.reset_index(drop=True)  
t2 = time.time()
print("处理完成，耗时：",t2-t1,"秒。")  
# 没有识别的记录，3条
hlhz6 = hlhz1.loc[hlhz1["hlmc"].isna()]
hlhz6 = ParseGPU(parser4, hlhz6, 3)   # 没有识别
hlhz6 = ParseGPU(parser3, hlhz6, 2)   # 没有识别
hlhz6 = ParseGPU(parser1, hlhz6, 0)   # 没有识别
hlhz6 = ParseGPU(parser2, hlhz6, 1)   # 识别

# 存盘
hlhz1.to_csv("D:/temp/data/增值税自开票2017/hlhz-hanlp-gpu-utf8-1k.csv",sep='~',encoding="utf-8", \
            index=False,header=None)  
